"""Tracking dataset exporters"""

from __future__ import annotations

import shutil
from collections import deque
from pathlib import Path

import networkx as nx
import numpy as np
import pandas as pd
import tifffile
from shapely.geometry import MultiPolygon
from tqdm.auto import tqdm

from acia.base import BaseImage, Contour, ImageSequenceSource, Overlay
from acia.tracking import TrackingSource
from acia.utils import multi_mask_to_polygons


class CellTrackingChallengeDatasetGT:
    """
    Utility class to create segmentation/tracking output in form of the Cell Tracking Challenge (CTC)
    """

    def __init__(self):
        self.sources = []

    def add(self, content: tuple[ImageSequenceSource, TrackingSource]):
        self.sources.append(content)

    def write(self, base_folder: str | Path = "data", offset=0):
        """Exports a ctc ground truth dataset

        Args:
            base_folder (str | Path, optional): _description_. Defaults to "data".
        """

        base_folder = Path(base_folder)

        base_folder.mkdir(exist_ok=True, parents=True)

        for i, (image_source, tracking_source) in enumerate(self.sources):

            mode = "GT"

            image_dir = base_folder / f"{i + offset:02}"
            ann_dir = base_folder / f"{i + offset:02}_{mode}"

            if image_dir.exists():
                shutil.rmtree(image_dir)
            if ann_dir.exists():
                shutil.rmtree(ann_dir)

            image_dir.mkdir()
            ann_dir.mkdir()

            height, width = -1, -1

            # save the images
            for t, image in enumerate(image_source):
                if isinstance(image, BaseImage):
                    image = image.raw
                height, width = image.shape[:2]
                tifffile.imwrite(str(image_dir / f"t{t:04}.tif"), image[..., 0])

            tracking_helper = CTCTrackingHelper(
                tracking_source.overlay, tracking_source.tracking_graph, height, width
            )
            ctc_masks, ctc_tracking_format = tracking_helper.to_ctc_format()

            # store segmentation masks
            seg_dir = ann_dir / "SEG"
            seg_dir.mkdir(exist_ok=True)

            for t, mask in enumerate(ctc_masks):
                tifffile.imwrite(str(seg_dir / f"man_seg{t:04}.tif"), mask)

            # store tracking masks
            track_dir = ann_dir / "TRA"
            track_dir.mkdir(exist_ok=True)

            for t, mask in enumerate(ctc_masks):
                tifffile.imwrite(str(track_dir / f"man_track{t:04}.tif"), mask)

            with open(
                str(track_dir / "man_track.txt"), "w", encoding="utf-8"
            ) as output_file:
                output_file.write("\n".join(ctc_tracking_format))


class CellTrackingDatasetResult:
    """
    Utility class to create segmentation/tracking output in form of the Cell Tracking Challenge (CTC)
    """

    def __init__(self):
        self.sources = []

    def add(self, content: tuple[TrackingSource, tuple[int, int]]):
        self.sources.append(content)

    def write(self, base_folder: str | Path = "data", offset=0):

        base_folder = Path(base_folder)

        base_folder.mkdir(exist_ok=True, parents=True)

        for i, (tracking_source, (height, width)) in enumerate(self.sources):

            mode = "RES"

            ann_dir = base_folder / f"{i+offset:02}_{mode}"

            if ann_dir.exists():
                shutil.rmtree(ann_dir)

            ann_dir.mkdir()

            tracking_helper = CTCTrackingHelper(
                tracking_source.overlay, tracking_source.tracking_graph, height, width
            )
            ctc_masks, ctc_tracking_format = tracking_helper.to_ctc_format()

            for t, mask in enumerate(ctc_masks):
                tifffile.imwrite(str(ann_dir / f"mask{t:04}.tif"), mask)

            with open(
                str(ann_dir / "res_track.txt"), "w", encoding="utf-8"
            ) as output_file:
                output_file.write("\n".join(ctc_tracking_format))


class CTCTrackingHelper:
    """Helper class for the CTC format generation"""

    def __init__(
        self, overlay: Overlay, tracking_graph: nx.DiGraph, height: int, width: int
    ):
        """Create a new tracking helper object

        Args:
            tracking_graph (nx.DiGraph): tracking graph consisting of contour nodes
        """
        self.contour_lookup = {cont.id: cont for cont in overlay}
        self.tracking_graph = tracking_graph
        # compute the life-cycles of individual cells
        self.life_cycles = CTCTrackingHelper.compute_life_cycles(self.tracking_graph)
        # create lookup (cont id --> life cycle index)
        self.life_cycle_lookup = CTCTrackingHelper.create_life_cycle_lookup(
            self.life_cycles
        )
        self.overlay = overlay
        self.height = height
        self.width = width

    def to_ctc_format(self) -> tuple[list[np.ndarray], list[str]]:
        """Generate CTC format

        Returns:
            Tuple[List[np.ndarray], List[str]]: Returns a List of segm/tracking masks and a list of lines for the ctc txt tracking format.
        """

        # generate txt format
        ctc_tracking_format = CTCTrackingHelper.txt_format(
            self.tracking_graph,
            self.life_cycles,
            self.contour_lookup,
            self.life_cycle_lookup,
        )

        ctc_masks = []
        # generate masks
        for overlay in self.overlay.timeIterator():
            mask = CTCTrackingHelper.convert_overlay_to_ctc_mask(
                overlay, self.life_cycle_lookup, self.height, self.width
            )
            ctc_masks.append(mask)

        return ctc_masks, ctc_tracking_format

    @staticmethod
    def compute_life_cycles(tracking_graph: nx.DiGraph) -> list[list[str]]:
        """Track life cycles of contour observations (from birth to division).

        Args:
            tracking_graph (nx.DiGraph): tracking graph

        Returns:
            List[List[str]]: List of life cycles (consisting of a list of contour ids)
        """

        start_nodes = deque(
            filter(lambda n: tracking_graph.in_degree(n) == 0, tracking_graph.nodes)
        )

        life_cycles = []

        while len(start_nodes) > 0:
            node = start_nodes.pop()
            life_cycle = [node]
            while tracking_graph.out_degree(node) == 1:
                node = next(tracking_graph.successors(node))
                life_cycle.append(node)

            life_cycles.append(life_cycle)

            if tracking_graph.out_degree(node) > 1:
                start_nodes += list(tracking_graph.successors(node))

        return life_cycles

    @staticmethod
    def txt_format(
        tracking_graph: nx.DiGraph,
        life_cycles: list[list[any]],
        contour_lookup,
        life_cycle_lookup,
    ) -> list[str]:
        """Generate CTC tracking txt format

        Args:
            tracking_graph (nx.DiGraph): tracking graph
            life_cycles (List[List[str]]): life_cycles

        Raises:
            Exception: when a life_cycle has multiple parents

        Returns:
            List[str]: list of ctc tracking lines
        """
        lines = []
        for i, life_cycle in enumerate(life_cycles):
            start_frame = contour_lookup[life_cycle[0]].frame
            end_frame = contour_lookup[life_cycle[-1]].frame

            # get the parent and its life_cycle number
            parents = list(tracking_graph.predecessors(life_cycle[0]))
            if len(parents) == 0:
                parent = 0  # TODO: default value when no parent???
            elif len(parents) == 1:
                parent = (
                    life_cycle_lookup[parents[0]] + 1
                )  # life-cycle enumeration starts with 1 #list(filter(lambda el: el[[-1] == parents[0], enumerate(life_cycles)))[0]
            else:
                raise Exception(
                    "Multiple parents not supported by Cell Tracking Challenge format!"
                )

            lines.append(
                f"{i+1} {start_frame} {end_frame} {parent}"
            )  # life-cycle enumeration starts with one

        return lines

    @staticmethod
    def create_life_cycle_lookup(life_cycles: list[list[Contour]]):
        """Computes a mapping from contour ids to life cycle ids

        Args:
            life_cycles (List[List[Contour]]): _description_

        Returns:
            _type_: _description_
        """
        contour_life_cycle_lookup = {}
        for i, life_cycle in enumerate(life_cycles):
            for cont_id in life_cycle:
                contour_life_cycle_lookup[
                    cont_id
                ] = i  # life cycle enumeration starts with 1

        return contour_life_cycle_lookup

    @staticmethod
    def convert_overlay_to_ctc_mask(
        overlay: Overlay,
        contour_life_cycle_lookup: dict[str, int],
        height: int,
        width: int,
    ) -> list[np.ndarray]:
        """Creates a ctc masks with correct numbering for a frame overlay

        Args:
            overlay (Overlay): overlay containing the contours
            contour_life_cycle_lookup (Dict[str, int]): lookup for the life cycle

        Returns:
            List[np.ndarray]: list of ctc masks
        """
        assert height > 0 and width > 0

        image_mask = np.zeros((height, width), dtype=np.uint16)
        for cont in overlay:
            life_cycle_id = (
                contour_life_cycle_lookup[cont.id] + 1
            )  # lifecycle ids must start with 1
            cont_mask = cont.toMask(height, width)
            image_mask = np.maximum(
                image_mask, (cont_mask * life_cycle_id).astype(np.uint16)
            )

        return image_mask

    @staticmethod
    def __load_masks(mask_path: Path) -> Overlay:
        """Load CTC masks into an overlay

        Args:
            mask_path (Path): Path to the folder containing the masks

        Returns:
            Overlay: Overlay containing all cell detections
        """

        mask_path = Path(mask_path)

        # load masks
        mask_files = sorted(mask_path.glob("man_seg*.tif"))
        masks = [tifffile.imread(file) for file in mask_files]

        all_polygons = map(
            multi_mask_to_polygons, tqdm(masks, desc="Convert masks to polygon...")
        )

        contours = []

        for frame, frame_polygons in enumerate(
            tqdm(all_polygons, desc="Convert to overlay...")
        ):
            for id, poly in frame_polygons:

                if isinstance(poly, MultiPolygon):
                    # if it comes to parsing problems take the polygon with the largest area
                    polygons = list(poly.geoms)
                    areas = [p.area for p in polygons]
                    poly = polygons[np.argmax(areas)]

                cc = np.stack(poly.exterior.coords.xy, axis=-1)
                contours.append(Contour(cc, -1, frame, f"{frame}_{id}"))

        return Overlay(contours, frames=list(range(0, len(masks))))

    @staticmethod
    def __load_tracking(tracking_path: Path) -> nx.DiGraph:
        """Load tracking from CTC txt file

        Args:
            tracking_path (Path): path to the folder containing the man_track.txt file

        Returns:
            nx.DiGraph: tracking graph with nodes and edges of every cell detection
        """

        track_annotation = pd.read_csv(
            tracking_path / "man_track.txt",
            delimiter=" ",
            names=["track_id", "start_frame", "end_frame", "parent_id"],
            header=None,
        )

        tracking_graph = nx.DiGraph()
        lc_lookup = {}

        for _, row in track_annotation.iterrows():
            start_frame = row["start_frame"]
            end_frame = row["end_frame"]
            id = row["track_id"]
            parent_id = row["parent_id"]

            node_items = [
                f"{frame}_{id}" for frame in range(start_frame, end_frame + 1)
            ]

            # add trajectory nodes
            for frame in range(start_frame, end_frame + 1):
                lc_lookup[f"{frame}_{id}"] = id
                tracking_graph.add_node(f"{frame}_{id}", frame=frame)

            # add trajectory edges
            for a, b in zip(node_items, node_items[1:]):
                tracking_graph.add_edge(a, b)

            # add edge to parent (0 means no parent)
            if parent_id != 0:
                tracking_graph.add_edge(
                    f"{start_frame - 1}_{parent_id}", f"{start_frame}_{id}"
                )

        return tracking_graph

    @staticmethod
    def load_ctc_format(
        segmentation_path: Path, tracking_path: Path
    ) -> tuple[Overlay, nx.DiGraph]:
        """Load ctc format from paths

        Args:
            segmentation_path (Path): path to the segmentation masks
            tracking_path (Path): path to the tracking gt/st

        Returns:
            tuple[Overlay, nx.DiGraph]: the overlay and tracking graph
        """

        # load segmentation overlay
        overlay = CTCTrackingHelper.__load_masks(mask_path=segmentation_path)

        # load tracking overlay
        tracking_graph = CTCTrackingHelper.__load_tracking(tracking_path=tracking_path)

        # Remove items only available in segmentation or tracking

        seg_ids = list(cont.id for cont in overlay)
        track_ids = set(tracking_graph.nodes())

        id_sym_difference = set(seg_ids).symmetric_difference(track_ids)

        # remove nodes in graph
        nodes_to_remove = set(tracking_graph.nodes()).intersection(id_sym_difference)
        for n in nodes_to_remove:
            tracking_graph.remove_node(n)

        # remove nodes in overlay
        conts_to_remove = set(seg_ids).intersection(id_sym_difference)
        for cont_id in conts_to_remove:
            seg_ids = np.array([cont.id for cont in overlay])
            index = np.argmin(seg_ids == cont_id)
            del overlay.contours[index]

        assert (
            len(
                {cont.id for cont in overlay}.symmetric_difference(
                    tracking_graph.nodes()
                )
            )
            == 0
        )

        return overlay, tracking_graph



"""Utilities for working with OMERO shape objects"""

from typing import List, Tuple

import omero
from omero.rtypes import rdouble, rint, rstring

# from omero.model.enums import UnitsLength


# Another helper for generating the color integers for shapes
def rgba_to_int(red, green, blue, alpha=255):
    """Return the color as an Integer in RGBA encoding"""
    r = red << 24
    g = green << 16
    b = blue << 8
    a = alpha
    rgba_int = r + g + b + a
    if rgba_int > (2**31 - 1):  # convert to signed 32-bit int
        rgba_int = rgba_int - 2**32
    return rgba_int


def create_rectangle(x: float, y: float, width: float, height: float, z: int, t: int):
    # create a rectangle shape (added to ROI below)
    print(
        f"Adding a rectangle at theZ: {z}, theT: {t}, X: {x}, Y: {y}, width: {width}, height: {height}"
    )
    rect = omero.model.RectangleI()
    rect.x = rdouble(x)
    rect.y = rdouble(y)
    rect.width = rdouble(width)
    rect.height = rdouble(height)
    rect.theZ = rint(z)
    rect.theT = rint(t)
    rect.textValue = rstring("test-Rectangle")
    rect.fillColor = rint(rgba_to_int(255, 255, 255, 255))
    rect.strokeColor = rint(rgba_to_int(255, 255, 0, 255))

    return rect


def create_ellipse(x: float, y: float, width: float, height: float, z: int, t: int):
    # create an Ellipse shape (added to ROI below)
    ellipse = omero.model.EllipseI()
    ellipse.x = rdouble(x)
    ellipse.y = rdouble(y)
    ellipse.radiusX = rdouble(width)
    ellipse.radiusY = rdouble(height)
    ellipse.theZ = rint(z)
    ellipse.theT = rint(t)
    ellipse.textValue = rstring("test-Ellipse")

    return ellipse


def make_polystr(points):
    return rstring(" ".join(f"{str(x)},{str(y)}" for x, y in points))


def make_coordinates(polystr: str) -> List[Tuple[float]]:
    return [tuple(map(float, textCoord.split(","))) for textCoord in polystr.split(" ")]


def create_polygon(
    points: List,
    z: int,
    t: int,
    fillColor=(255, 0, 255, 50),
    strokeColor=(255, 255, 0),
    description="",
):
    # create an ROI with a single polygon, setting colors and lineWidth
    polygon = omero.model.PolygonI()
    polygon.theZ = rint(z)
    polygon.theT = rint(t)
    polygon.fillColor = rint(rgba_to_int(*fillColor))
    polygon.strokeColor = rint(rgba_to_int(*strokeColor))
    # polygon.strokeWidth = omero.model.LengthI(10, UnitsLength.PIXEL)
    polygon.points = make_polystr(points)
    polygon.textValue = rstring(description)

    return polygon



""" Utils for OMERO segmenation data"""

from typing import List

import cv2
import numpy as np
import omero
from omero.gateway import BlitzGateway, DatasetWrapper, ImageWrapper, ProjectWrapper
from PIL import Image, ImageDraw, ImageFont

from acia.segm.omero.storer import OmeroSequenceSource
from acia.utils import ScaleBar


def getImage(conn: BlitzGateway, imageId: int) -> ImageWrapper:
    """Get omero image by id
        Note: only images in your current group are accessible

    Args:
        conn (BlitzGateway): current omero connection
        imageId (int): image id

    Returns:
        ImageWrapper: image object
    """
    return conn.getObject("Image", imageId)


def getDataset(conn: BlitzGateway, datasetId: int) -> DatasetWrapper:
    """Get omero dataset by id
        Note: only datasets in your current group are accessible

    Args:
        conn (BlitzGateway): active omero connection
        datasetId (int): dataset id

    Returns:
        DatasetWrapper: dataset object
    """
    return conn.getObject("Dataset", datasetId)


def getProject(conn: BlitzGateway, projectId: int) -> ProjectWrapper:
    """Get omero project by id
        Note: only projects in your current group are accessible

    Args:
        conn (BlitzGateway): active omero connection
        projectId (int): project id

    Returns:
        ProjectWrapper: project object
    """
    return conn.getObject("Project", projectId)


def list_projects(conn: BlitzGateway) -> List[ProjectWrapper]:
    """List projects in the current user group
        Note: only projects in your current group are accessible

    Args:
        conn (BlitzGateway): Current omero BlitzGateway connection

    Returns:
        List[ProjectWrapper]: List of project wrappers
    """
    return conn.getObjects("Project")


def list_image_ids_in_dataset(conn: BlitzGateway, datasetId: int) -> List[int]:
    """[summary]

    Args:
        conn (BlitzGateway): active omero connection
        datasetId (int): dataset id

    Returns:
        List[int]: array of all image ids of the dataset
    """
    return [
        image.getId() for image in conn.getObjects("Image", opts={"dataset": datasetId})
    ]


def list_images_in_dataset(conn: BlitzGateway, datasetId: int) -> List[ImageWrapper]:
    """List all images in the omero dataset

    Args:
        conn (BlitzGateway): active omero connection
        datasetId (int): dataset id

    Returns:
        List[ImageWrapper]: List of omero images
    """
    return conn.getObjects("Image", opts={"dataset": datasetId})


def list_datasets_in_project(
    conn: BlitzGateway, projectId: int
) -> List[DatasetWrapper]:
    return conn.getObjects("Dataset", opts={"project": projectId})


def list_images_in_project(conn: BlitzGateway, projectId: int) -> List[ImageWrapper]:
    return [
        image
        for dataset in list_datasets_in_project(conn, projectId=projectId)
        for image in dataset.listChildren()
    ]


def list_image_ids_in(omero_id: int, omero_type: str, conn: BlitzGateway) -> List[int]:
    """List all image sequences in an omero source (dataset, project or image)

    Args:
        omero_id (int): the omero id specifying the resource on the omero server
        omero_type (str): the type of the omero source. Can be 'project', 'dataset' or 'image'.
        conn (BlitzGateway): the connection to omero

    Raises:
        Exception: when wrong omero_type is specified

    Returns:
        List[int]: list of image ids contained in the OMERO resource
    """

    omero_type = omero_type.lower()
    func = None

    if omero_type == "project":
        func = list_images_in_project
    elif omero_type == "dataset":
        func = list_images_in_dataset
    elif omero_type == "image":
        return [omero_id]
    else:
        raise Exception(
            f"Wrong omero_type: '{omero_type}'! Please choose one of 'project', 'dataset' or 'image'!"
        )

    return list(map(lambda image: image.getId(), func(conn, omero_id)))


def get_image_name(conn: BlitzGateway, imageId: int) -> str:
    return conn.getObject("Image", imageId).getName()


def get_project_name(conn: BlitzGateway, projectId: int) -> str:
    return conn.getObject("Project", projectId).getName()


def image_iterator(conn: BlitzGateway, object) -> ImageWrapper:
    if object.OMERO_CLASS == "Image":
        yield object
    if object.OMERO_CLASS == "Dataset":
        yield from list_images_in_dataset(conn, object.getId())
    if object.OMERO_CLASS == "Project":
        yield from list_images_in_project(conn, object.getId())


def create_project(conn: BlitzGateway, project_name: str) -> ProjectWrapper:
    new_project = ProjectWrapper(conn, omero.model.ProjectI())
    new_project.setName(project_name)
    new_project.save()
    return new_project


def create_dataset(
    conn: BlitzGateway, projectId: int, dataset_name: str
) -> DatasetWrapper:
    # Use omero.gateway.DatasetWrapper:
    new_dataset = DatasetWrapper(conn, omero.model.DatasetI())
    new_dataset.setName(dataset_name)
    new_dataset.save()
    # Can get the underlying omero.model.DatasetI with:
    dataset_obj = new_dataset._obj

    # Create link to project
    link = omero.model.ProjectDatasetLinkI()
    # We can use a 'loaded' object, but we might get an Exception
    # link.setChild(dataset_obj)
    # Better to use an 'unloaded' object (loaded = False)
    link.setChild(omero.model.DatasetI(dataset_obj.id.val, False))
    link.setParent(omero.model.ProjectI(projectId, False))
    conn.getUpdateService().saveObject(link)

    return new_dataset


class OmeroScaleBar(ScaleBar):
    """Renderer for scale bar to show metric size of pixels on image"""

    def __init__(
        self,
        oss: OmeroSequenceSource,
        width,
        unit="MICROMETER",
        short_title="μm",
        color=(255, 255, 255),
        font_size=25,
    ):
        self.width = width
        self.unit = unit
        self.color = color
        self.short_title = short_title
        self.font_size = font_size

        pixelSizes = oss.rawPixelSize

        pixelSize = omero.model.LengthI(pixelSizes[0], self.unit).getValue()

        self.pixelWidth = int(np.round(self.width / pixelSize))

        # TODO: make parameter
        self.pixelHeight = 10

        # TODO: no fixed font file
        self.font = ImageFont.truetype(
            "/usr/share/fonts/truetype/freefont/FreeSans.ttf", self.font_size
        )

    def draw(self, image, xstart, ystart):
        # TODO: thickness parameter
        # draw line
        cv2.line(
            image, (xstart, ystart), (xstart + self.pixelWidth, ystart), self.color, 1
        )
        half_y = int(np.round(self.pixelHeight / 2))
        cv2.line(
            image, (xstart, ystart - half_y), (xstart, ystart + half_y), self.color
        )
        cv2.line(
            image,
            (xstart + self.pixelWidth, ystart - half_y),
            (xstart + self.pixelWidth, ystart + half_y),
            self.color,
        )

        # draw size
        unit_text = self.short_title
        text = f"{self.width} {unit_text}"
        img_pil = Image.fromarray(image)
        draw = ImageDraw.Draw(img_pil)
        text_x, text_y = draw.textsize(text, font=self.font)
        draw.text(
            (xstart + self.pixelWidth / 2 - text_x / 2, ystart - text_y - 2),
            text,
            fill=self.color,
            font=self.font,
        )
        image = np.array(img_pil)

        return image


def has_all_tags(object, tag_list: List[str] = None):
    if tag_list is None:
        tag_list = []

    tag_list = tag_list.copy()
    for ann in object.listAnnotations():
        if ann.OMERO_TYPE == omero.model.TagAnnotationI:
            if ann.getTextValue() in tag_list:
                del tag_list[tag_list.index(ann.getTextValue())]

        if len(tag_list) == 0:
            break

    return len(tag_list) == 0



"""Top-level package for AutomatedCellularImageAnalysis."""

__author__ = """Johannes Seiffarth"""
__email__ = "j.seiffarth@fz-juelich.de"
__version__ = "0.2.37"


from pint import UnitRegistry

ureg = UnitRegistry()
Q_ = ureg.Quantity
U_ = ureg.Unit



""" Testcases for single-cell property extractors """

import unittest
from itertools import product

import numpy as np

from acia import ureg
from acia.analysis import (
    AreaEx,
    DynamicTimeEx,
    ExtractorExecutor,
    FluorescenceEx,
    FrameEx,
    IdEx,
    LengthEx,
    PositionEx,
    PropertyExtractor,
    TimeEx,
    WidthEx,
)
from acia.base import Contour, Overlay
from acia.segm.local import InMemorySequenceSource, LocalImageSource


class TestPropertyExtractors(unittest.TestCase):
    """Test cases for single-cell property extractors"""

    def test_unit_conversion(self):
        # test basic conversion patterns

        self.assertAlmostEqual(
            PropertyExtractor("test", "meter", "millimeter").convert(1), 1000
        )
        self.assertAlmostEqual(
            PropertyExtractor("test", "micrometer", "millimeter").convert(1), 1e-3
        )
        self.assertAlmostEqual(
            PropertyExtractor("test", "liter", "milliliter").convert(1), 1000
        )
        self.assertAlmostEqual(
            PropertyExtractor("test", "micrometer", "micrometer").convert(1), 1
        )
        self.assertAlmostEqual(
            PropertyExtractor("test", "meter", "micrometer").convert(1), 1e6
        )

    def test_extractors(self):
        # in x,y coordinates
        contours = [Contour([[0, 0], [2, 0], [2, 3], [0, 3]], -1, frame=0, id=23)]
        overlay = Overlay(contours)

        image = np.zeros((200, 200))
        image[0, 0] = 2
        image[0, 1] = 5
        image[1, 0] = 6
        image[1, 1] = 10
        image[2, 0:3] = 4
        image[0:2, 2] = 4
        image_source = LocalImageSource.from_array(image)

        # pixel size
        ps = 0.07

        # test basic extractors
        df = ExtractorExecutor().execute(
            overlay=overlay,
            images=image_source,
            extractors=[
                IdEx(),
                FrameEx(),
                AreaEx(input_unit=(ps * ureg.micrometer) ** 2),
                LengthEx(input_unit=ps * ureg.micrometer),
                WidthEx(input_unit=ps * ureg.micrometer),
                TimeEx(input_unit="15 * minute"),  # one frame every 15 minutes
                PositionEx(input_unit=ps * ureg.micrometer),
                FluorescenceEx(channels=[0], channel_names=["gfp"], parallel=1),
                FluorescenceEx(
                    channels=[0],
                    channel_names=["gfp_mean"],
                    summarize_operator=np.mean,
                    parallel=1,
                ),
            ],
        )

        self.assertEqual(df["area"][0], (2 * 3) * ps**2)
        self.assertEqual(df["length"][0], 3 * ps)
        self.assertEqual(df["id"][0], 23)
        self.assertEqual(df["frame"][0], 0)
        self.assertEqual(df["time"][0], 0 * 15 / 60)
        self.assertEqual(df["position_x"][0], 2 / 2 * ps)
        self.assertEqual(df["position_y"][0], 3 / 2 * ps)
        self.assertEqual(df["gfp"][0], np.median(image[:3, :2]))
        self.assertEqual(df["gfp_mean"][0], np.mean(image[:3, :2]))

    def test_dynamic_time_extractor(self):
        # in x,y coordinates
        contours = [
            Contour([[0, 0], [2, 0], [2, 3], [0, 3]], -1, frame=0, id=23),
            Contour([[0, 0], [2, 0], [2, 3], [0, 3]], -1, frame=1, id=24),
            Contour([[0, 0], [2, 0], [2, 3], [0, 3]], -1, frame=2, id=25),
        ]
        overlay = Overlay(contours)

        timepoints = [1710326746.8015938, 1710326987.554663, 1710327228.3607492]
        rel_timepoints = np.array(timepoints) - timepoints[0]

        df = ExtractorExecutor().execute(
            overlay=overlay,
            images=[LocalImageSource(None)] * 3,
            extractors=[IdEx(), FrameEx(), DynamicTimeEx(timepoints, relative=True)],
        )

        self.assertEqual(df["time"][0], 0)
        self.assertEqual(df["time"][1], rel_timepoints[1] / 3600)
        self.assertEqual(df["time"][2], rel_timepoints[2] / 3600)

    def test_dynamic_time_extractor_failures(self):
        contours = [
            Contour([[0, 0], [2, 0], [2, 3], [0, 3]], -1, frame=0, id=23),
            Contour([[0, 0], [2, 0], [2, 3], [0, 3]], -1, frame=1, id=24),
            Contour([[0, 0], [2, 0], [2, 3], [0, 3]], -1, frame=2, id=25),
        ]
        overlay = Overlay(contours)

        with self.assertRaises(ValueError) as _:
            DynamicTimeEx([])

        with self.assertRaises(ValueError) as _:
            _ = ExtractorExecutor().execute(
                overlay=overlay,
                images=[LocalImageSource(None)] * 3,
                extractors=[
                    IdEx(),
                    FrameEx(),
                    DynamicTimeEx(timepoints=[1, 2], relative=True),
                ],
            )

    def test_fluorescence_extractor_float(self):
        """Testing that the fluorescence exporter can work with float values"""

        contours = [Contour([[0, 0], [2, 0], [2, 3], [0, 3]], -1, frame=0, id=23)]
        overlay = Overlay(contours)

        image = np.zeros((200, 200), dtype=np.float32)
        image[0, 0] = 2.5
        image[0, 1] = 5.5
        image[1, 0] = 6
        image[1, 1] = 10.1
        image[2, 0:3] = 4
        image[0:2, 2] = -4.3
        image_source = LocalImageSource.from_array(image)

        # test basic extractors
        df = ExtractorExecutor().execute(
            overlay=overlay,
            images=image_source,
            extractors=[
                FluorescenceEx(channels=[0], channel_names=["gfp"], parallel=1),
                FluorescenceEx(
                    channels=[0],
                    channel_names=["gfp_mean"],
                    summarize_operator=np.mean,
                    parallel=1,
                ),
            ],
        )

        self.assertEqual(df["gfp"][0], np.median(image[:3, :2]))
        self.assertEqual(df["gfp_mean"][0], np.mean(image[:3, :2]))

    def test_parallel_fluorescence_extraction(self):
        squared_num = 30
        contours = [
            Contour([[0, 0], [2, 0], [2, 2], [0, 2]], -1, frame=frame, id=id)
            for id, frame in product(list(range(squared_num)), list(range(squared_num)))
        ]
        overlay = Overlay(contours)

        image = np.zeros((200, 200))
        image[0, 0] = 2
        image[0, 1] = 5
        image[1, 0] = 6
        image[1, 1] = 10
        image_sources = InMemorySequenceSource(np.stack([image] * squared_num))

        self.assertTrue(image_sources is not None)

        # test basic extractors
        df = ExtractorExecutor().execute(
            overlay=overlay,
            images=image_sources,
            extractors=[
                FluorescenceEx(channels=[0], channel_names=["fl1"]),
                FluorescenceEx(
                    channels=[0], channel_names=["fl1_mean"], summarize_operator=np.mean
                ),
            ],
        )

        np.testing.assert_array_equal(df["fl1"], [5.5] * len(df))
        np.testing.assert_array_equal(
            df["fl1_mean"], [np.mean([2, 5, 6, 10])] * len(df)
        )


if __name__ == "__main__":
    unittest.main()



"""Unit test cases for tracking subsampling"""

import unittest
from typing import List

import networkx as nx
import numpy as np

from acia.base import Contour, Overlay
from acia.tracking import TrackingSource, TrackingSourceInMemory, subsample_tracking
from acia.tracking.utils import delete_nodes, subsample_lineage


class TestSubsampling(unittest.TestCase):
    """Test the linearization of z and t stacks"""

    @staticmethod
    def linear_track(
        length, frame_start, id_start=None, successors: List[TrackingSource] = None
    ):

        if successors is None:
            successors = []

        if id_start is None:
            id_start = 0
            for succ in successors:
                id_start = np.max([id_start, *[cont.id for cont in succ.overlay]]) + 1

        ids = [id_start + i for i in range(length)]
        frames = [frame_start + i for i in range(length)]

        overlay = Overlay(
            [Contour(None, -1, frame, id) for id, frame in zip(ids, frames)]
        )
        tracking_graph = nx.DiGraph()
        tracking_graph.add_nodes_from([cont.id for cont in overlay])
        tracking_graph.add_edges_from(
            [(a.id, b.id) for a, b in zip(overlay.contours, overlay.contours[1:])]
        )

        linear_ts = TrackingSourceInMemory(overlay, tracking_graph)

        for succ in successors:
            # merge the two trackings
            linear_ts.merge(succ)

            # add edge between last and first overlay contour
            linear_ts.tracking_graph.add_edge(
                overlay.contours[-1].id, succ.overlay.contours[0].id
            )

        return linear_ts

    def test_linear(self):
        """Test subsampling of a non-dividing contour"""

        # create tracking source
        tsim = TestSubsampling.linear_track(10, 0)

        # subsample tracking
        tr_source = subsample_tracking(tsim, 2)

        self.assertEqual(tr_source.tracking_graph.number_of_nodes(), 5)
        self.assertEqual(tr_source.tracking_graph.number_of_edges(), 4)

        self.assertSetEqual(
            set(map(lambda c: c.id, tr_source.overlay)), {0, 2, 4, 6, 8}
        )

    def test_division(self):
        """Test subsampling of a dividing contour"""

        # create tracking source
        tsim = TestSubsampling.linear_track(
            5,
            0,
            successors=[
                TestSubsampling.linear_track(3, 5, id_start=0),
                TestSubsampling.linear_track(5, 5, id_start=3),
            ],
        )

        # subsample tracking
        tr_source = subsample_tracking(tsim, 3)

        self.assertEqual(tr_source.tracking_graph.number_of_nodes(), 5)
        self.assertEqual(tr_source.tracking_graph.number_of_edges(), 4)
        self.assertSetEqual(
            set(map(lambda c: c.id, tr_source.overlay)), {8, 11, 1, 4, 7}
        )
        self.assertEqual(tr_source.tracking_graph.out_degree(11), 2)


def linear_track(start_frame: int, num_frames: int) -> nx.DiGraph:
    """Creates a tracklet for a single cell

    Args:
        start_frame (int): start frame
        num_frames (int): number of frames

    Returns:
        nx.DiGraph: Simple lineage with a single cell track (frame is node attribute)
    """

    lineage = nx.DiGraph()
    nodes = list(range(start_frame, start_frame + num_frames))

    lineage.add_nodes_from(nodes)
    for n in lineage.nodes:
        lineage.nodes[n]["frame"] = n

    for a, b in zip(nodes, nodes[1:]):
        lineage.add_edge(a, b)

    return lineage


class TestNewSubsampling(unittest.TestCase):
    """Test the linearization of z and t stacks"""

    def test_node_deletion(self):
        graph = nx.DiGraph()
        graph.add_nodes_from([1, 2, 3])
        graph.add_edges_from([(1, 2), (2, 3)])

        new_graph = delete_nodes(graph, [2])

        self.assertEqual(set(new_graph.nodes), {1, 3})
        self.assertEqual(set(new_graph.edges), {(1, 3)})

    def test_node_subsampling(self):
        lineage = linear_track(0, 11)

        sub_lineage = subsample_lineage(lineage, 5)

        self.assertEqual(set(sub_lineage.nodes), {0, 5, 10})
        self.assertEqual(set(sub_lineage.edges), {(0, 5), (5, 10)})

    def test_node_sub_with_div(self):
        lineage = nx.DiGraph()

        ########
        #           | 3 - 4 - 5
        # 0 - 1 - 2 - 6 - 7 - 8
        # 0 | 1 | 2 | 3 | 4 | 5 (frame index)

        # add nodes with frame information
        lineage.add_nodes_from(
            [
                (0, dict(frame=0)),
                (1, dict(frame=1)),
                (2, dict(frame=2)),
                (3, dict(frame=3)),
                (4, dict(frame=4)),
                (5, dict(frame=5)),
                (6, dict(frame=3)),
                (7, dict(frame=4)),
                (8, dict(frame=5)),
            ],
        )

        # add edges
        lineage.add_edges_from(
            [(0, 1), (1, 2), (2, 3), (3, 4), (4, 5), (2, 6), (2, 7), (2, 8)]
        )

        # subsample lineage
        new_graph = subsample_lineage(lineage, 2)

        # check nodes and connectivity
        self.assertEqual({0, 2, 4, 7}, set(new_graph.nodes))
        self.assertEqual({(0, 2), (2, 4), (2, 7)}, set(new_graph.edges))


if __name__ == "__main__":
    unittest.main()



""" Helper file to generate contours from masks"""

import logging
from typing import List

import cv2
import mmcv
import numpy as np
import rtree
import torch
from mmdet.apis import inference_detector
from shapely.geometry import LineString, Polygon

logger = logging.getLogger(__name__)


def contour_from_mask(mask, score_threshold):
    """
    Estimate largest contour from pixel-wise mask
    """
    contours, _ = cv2.findContours(
        np.where(mask > score_threshold, 1, 0).astype(np.uint8),
        cv2.RETR_TREE,
        cv2.CHAIN_APPROX_SIMPLE,
    )

    # select largest contour
    selected_contour = []
    for cont in contours:
        if len(cont) > len(selected_contour):
            selected_contour = cont

    return np.squeeze(selected_contour)


def prepare_contours(
    segm_result, labels, offset_x=0, offset_y=0, seg_score_threshold=0.3
):
    offset = np.array([offset_x, offset_y])

    all_contours = []
    if segm_result is not None and len(labels) > 0:  # non empty
        segms = mmcv.concat_list(segm_result)

        for seg in segms:
            seg = seg.astype(np.float32)

            # Creating kernel
            # kernel = np.ones((3, 3), np.uint8)
            # find contours with cv2
            contours, _ = cv2.findContours(
                np.where(seg > seg_score_threshold, 1, 0).astype(np.uint8),
                cv2.RETR_TREE,
                cv2.CHAIN_APPROX_SIMPLE,
            )

            selected_contour = []
            for cont in contours:
                if len(cont) > len(selected_contour):
                    selected_contour = cont

            cont_data = [
                (np.ceil(cont).astype(np.int32).squeeze(axis=1) + offset[None, :])
                for cont in [selected_contour]
                if len(cont) > 0
            ]  # [[str(x), str(y)] for cont in contours for x,y in cont]

            logging.info("Num contours: %d", len(contours))

            all_contours.append(cont_data)

    return all_contours


def postprocess(output_data, model, offset_x=0, offset_y=0, contours=False):
    segm_result = None

    if model.with_mask:
        bbox_result, segm_result = output_data
    else:
        bbox_result = output_data

    output_contours = (segm_result is not None) and contours

    bboxes = np.vstack(bbox_result)

    # print("Num bounding boxes:")

    # identify the labels of the detected boxes
    labels = [
        np.full(bbox.shape[0], i, dtype=np.int32) for i, bbox in enumerate(bbox_result)
    ]
    labels = np.concatenate(labels)

    # prepare contours if necessary
    if output_contours:
        contours = prepare_contours(segm_result, labels, offset_x, offset_y)
        logging.info("Num contours: %d", len(contours))
    else:
        contours = [[]] * len(bboxes)

    result = []

    for box, label, contour, segm in zip(
        bboxes, labels, contours, mmcv.concat_list(segm_result)
    ):
        if output_contours and len(contour) == 0:
            # skip this detection
            continue

        box_coords = list(map(float, box[:4]))
        score = float(box[4])
        label_name = model.CLASSES[label]

        # print("Class '%s' detected at %s with score %.2f" % (label_name, box_coords, score))

        result_dict = {
            "label": label_name,
            "bbox": box_coords,
            "score": score,
            "mask": segm,
        }

        # sort descending
        if output_contours:
            result_dict["contours"] = sorted(
                ({"x": cont[:, 0], "y": cont[:, 1]} for cont in contour),
                key=lambda cont: -len(cont["x"]),
            )

        result.append(result_dict)  # 'contours': contour

    return result


def tile_touch_filter(detection, image_tile_poly: LineString, threshold=10):
    """
    returns False iff detections it too close (<threshold) to the boundaries and likely to be restricted by them

    detection: the detection dict
    image_tile_poly: Line string of the image tile boundaries
    threshold: minimum allowed distance to the boundaries
    """

    contour = detection["contours"][0]
    detection_poly = Polygon(zip(contour["x"], contour["y"]))

    distance = detection_poly.distance(image_tile_poly)

    # print(distance)

    return distance >= threshold


def inference(image, model, offset_x=0, offset_y=0):
    # inference on tile
    raw_tile_results = inference_detector(model, image)

    # postprocess
    tile_results = postprocess(
        raw_tile_results, model, offset_x=offset_x, offset_y=offset_y
    )

    return tile_results


# TODO: make this function shorter
# pylint: disable=R0915
def tiled_inference(
    image,
    model,
    x_shift=256 - 128,
    y_shift=256 - 128,
    tile_width=256,
    tile_height=256,
    pd=25,
):
    """
    Execute inference in a tiled fashion

    x_shift: shift on the x-axis for every image slot
    y_shift: shift on the y-axis for every image slot
    tile_width: width of the image tile
    tile_height: height of the image tile

    TODO: When tiles align with image borders, we should not do tile touch filtering
    """

    # get the image dimensions
    height, width = image.shape[:2]

    x_start = 0
    y_start = 0

    all_detections = []

    # padding the image (rgb)
    padding_size = pd
    padded_image = np.zeros(
        (height + 2 * padding_size, width + 2 * padding_size, 3), dtype=np.uint8
    )
    padded_image[
        padding_size : padding_size + height, padding_size : padding_size + width
    ] = image
    orig_image = image
    image = padded_image

    height, width = image.shape[:2]

    # iterate over top coordinate of tile
    ys = list(range(max(1, 1 + int(np.ceil((height - tile_height) / y_shift)))))
    for iY in ys:
        y = y_start + iY * y_shift
        # iterate over left coordinate of tile
        xs = list(range(max(1, 1 + int(np.ceil((width - tile_width) / x_shift)))))
        for iX in xs:
            x = x_start + iX * x_shift

            # print(x,y)

            # compute the lower right coordinates of the tile
            y_end = min(height, y + tile_height)
            x_end = min(width, x + tile_width)

            # print(x_end, y_end)
            # print(y_end-y, x_end -x)
            # get the image tile
            image_tile = image[y:y_end, x:x_end]
            # print(image_tile.shape)
            # zero padding to constant tile size (otherwise we get devision errors)
            const_tile_format = np.zeros((tile_height, tile_width, 3), dtype=np.uint8)
            const_tile_format[: y_end - y, : x_end - x] = image_tile

            # print(const_tile_format.shape)

            tile_results = inference(const_tile_format, model, x, y)

            tile_results = list(
                filter(lambda det: np.sum(det["mask"]) >= 3, tile_results)
            )

            if len(tile_results) > 0:
                filter_mask = mask_nms(
                    np.stack([det["mask"] for det in tile_results]),
                    np.stack([det["bbox"] for det in tile_results]),
                    np.stack([det["score"] for det in tile_results]),
                )

                tile_results = list(np.array(tile_results)[filter_mask])

            # print(len(tile_results))

            # polygon for the image tile
            # image_tile_poly = LineString([(x, y), (x+tile_width, y), (x+tile_width, y+tile_height), (x, y+tile_height), (x,y)])

            # filter the detections
            #   -> no detections close to the border of the tile schould be considered
            filter_mask = np.ones(len(tile_results), dtype=bool)

            for i, det in enumerate(tile_results):
                det_mask = det["mask"]

                row, col = np.nonzero(det_mask)

                miny = np.min(row)
                maxy = np.max(row)
                minx = np.min(col)
                maxx = np.max(col)

                min_distance = np.min(
                    [miny, minx, tile_height - maxy, tile_width - maxx]
                )

                if min_distance < padding_size:
                    filter_mask[i] = False

            tile_results = list(np.array(tile_results)[filter_mask])
            # tile_results += filter(partial(tile_touch_filter, image_tile_poly=image_tile_poly), tile_results)

            new_masks = np.zeros((len(tile_results), *orig_image.shape[:2]), dtype=bool)

            # expand masks to full image
            for i, det in enumerate(tile_results):
                new_mask = new_masks[i]

                y_offset = 0
                y_endset = 0
                if iY == 0:
                    y_offset = pd
                if y_end > orig_image.shape[0] + pd:
                    y_endset = y_end - (orig_image.shape[0] + pd)
                x_offset = 0
                x_endset = 0
                if iX == 0:
                    x_offset = pd
                if x_end > orig_image.shape[1] + pd:
                    x_endset = x_end - (orig_image.shape[1] + pd)

                mask_height = (y_end - y) - y_endset
                mask_width = (x_end - x) - x_endset

                new_mask[
                    max(0, y - pd) : y - pd + mask_height,
                    max(0, x - pd) : x - pd + mask_width,
                ] = det["mask"][
                    y_offset:mask_height, x_offset:mask_width
                ]  # [:y_end - y,:x_end - x]
                det["mask"] = new_mask
                det["bbox"] += np.array([x, y, x, y]) - pd

            all_detections += tile_results

    return all_detections


def non_max_supression(all_detections: List[Polygon], iou=0.3):
    """
    Performing something like non-maximum supression on a list of detections

    TODO: make sure that this corresponds with some paper for nms

    all_detections: all detections found in an image
    iou: intersection over union: if a poly intersects more than that with another poly and it's score is lower it gets discarded.

    returns the filtered list of detections
    """
    # descending sort
    all_detections = sorted(all_detections, key=lambda det: det["score"])
    polygons = []
    for det in all_detections:
        contour = det["contours"][0]
        xs = contour["x"]
        ys = contour["y"]

        poly = Polygon(zip(xs, ys))

        if not poly.is_valid:
            logging.warning("Invalid polygon!")

        polygons.append(poly)

    idx = rtree.index.Index()
    for pos, poly in enumerate(polygons):
        idx.insert(pos, poly.bounds)

    set_remove_indices = set()

    # Loop through each Shapely polygon
    for i, poly in enumerate(polygons):
        score = all_detections[i]["score"]
        area = poly.area

        # Merge cells that have overlapping bounding boxes
        for pos in idx.intersection(poly.bounds):
            if pos == i:
                continue

            poly_other = polygons[pos]
            score_other = all_detections[pos]["score"]

            # distance = poly.distance(poly_other)
            # intersect = poly.intersects(poly_other)

            # print(poly.is_valid)
            # print(poly_other.is_valid)

            poly_other = poly_other.buffer(0)
            # print(poly_other.is_valid)

            # compute intersection
            intersect_poly = poly.intersection(poly_other)
            intersect_area = intersect_poly.area

            if score < score_other and intersect_area / area > iou:
                # do not take poly
                set_remove_indices.add(i)
                break

    return list(
        map(
            lambda idet: idet[1],
            filter(
                lambda idet: not idet[0] in set_remove_indices,
                enumerate(all_detections),
            ),
        )
    )


def np_vec_no_jit_iou(boxes1, boxes2):
    x11, y11, x12, y12 = np.split(boxes1, 4, axis=1)
    x21, y21, x22, y22 = np.split(boxes2, 4, axis=1)

    xA = np.maximum(x11, np.transpose(x21))
    yA = np.maximum(y11, np.transpose(y21))
    xB = np.minimum(x12, np.transpose(x22))
    yB = np.minimum(y12, np.transpose(y22))

    interArea = np.maximum((xB - xA + 1), 0) * np.maximum((yB - yA + 1), 0)
    boxAArea = (x12 - x11 + 1) * (y12 - y11 + 1)
    boxBArea = (x22 - x21 + 1) * (y22 - y21 + 1)
    iou = interArea / (boxAArea + np.transpose(boxBArea) - interArea)

    return iou


def torch_vec_no_jit_iou(boxes1, boxes2):
    x11, y11, x12, y12 = torch.chunk(boxes1, 4, dim=1)
    x21, y21, x22, y22 = torch.chunk(boxes2, 4, dim=1)

    xA = torch.maximum(x11, torch.transpose(x21, 0, 1))
    yA = torch.maximum(y11, torch.transpose(y21, 0, 1))
    xB = torch.minimum(x12, torch.transpose(x22, 0, 1))
    yB = torch.minimum(y12, torch.transpose(y22, 0, 1))

    interArea = torch.maximum(
        (xB - xA + 1), torch.tensor(0, device=boxes1.device)
    ) * torch.maximum((yB - yA + 1), torch.tensor(0, device=boxes1.device))
    boxAArea = (x12 - x11 + 1) * (y12 - y11 + 1)
    boxBArea = (x22 - x21 + 1) * (y22 - y21 + 1)
    iou = interArea / (boxAArea + torch.transpose(boxBArea, 0, 1) - interArea)

    return iou


def torch_mask_nms(
    masks,
    bboxes,
    scores,
    bbox_iou_threshold=0.1,
    mask_iou_threshold=0.4,
    score_threshold=0.1,
):
    """
    iou: if intersection between two cells is larger, only take the better scored one
    """

    device = "cuda:0"

    masks = torch.tensor(masks, device=device)
    bboxes = torch.tensor(bboxes, device=device)

    # areas = torch.sum(torch.tensor(masks), axis=(1,2)).numpy()

    bbox_iou = torch_vec_no_jit_iou(bboxes, bboxes)

    print(masks.shape)

    # print(masks.nbytes)

    filter_mask = scores >= score_threshold

    scores = torch.tensor(scores, device=device)

    drops = []  # torch.zeros_like(scores, dtype=torch.bool)

    # intersection = masks[None] & np.r
    for i, (mask, score) in enumerate(zip(masks, scores)):
        if not filter_mask[i]:
            continue

        candidate_mask = bbox_iou[i, :] > bbox_iou_threshold

        # area = areas[i]#np.sum(mask)
        intersection = mask[None, :] & masks[candidate_mask]
        joint = mask[None, :] | masks[candidate_mask]

        intersection_areas = torch.sum(intersection, dim=(1, 2))
        joint_areas = torch.sum(joint, dim=(1, 2))

        relative_intersections = intersection_areas / joint_areas

        over_threshold = torch.where(relative_intersections > mask_iou_threshold)

        higher_scored = scores[candidate_mask][over_threshold] > score

        drops.append(~(torch.sum(higher_scored) >= 1).cpu())

        # if drop:
        #    filter_mask[i] = False

        # print(relative_intersections)

    filter_mask = np.array(drops, dtype=bool)

    return np.arange(len(masks))[filter_mask]


def mask_nms(
    masks,
    bboxes,
    scores,
    bbox_iou_threshold=0.1,
    mask_iou_threshold=0.4,
    score_threshold=0.1,
):
    """
    iou: if intersection between two cells is larger, only take the better scored one
    """
    # masks = torch.tensor(masks)
    # bboxes = torch.tensor(bboxes)

    # areas = torch.sum(torch.tensor(masks), axis=(1, 2)).numpy()

    bbox_iou = torch_vec_no_jit_iou(torch.tensor(bboxes), torch.tensor(bboxes)).numpy()

    print(masks.shape)

    print(masks.nbytes)

    filter_mask = scores >= score_threshold
    # intersection = masks[None] & np.r
    for i, (mask, score) in enumerate(zip(masks, scores)):
        if not filter_mask[i]:
            continue

        candidate_mask = bbox_iou[i, :] > bbox_iou_threshold

        # area = areas[i]#np.sum(mask)
        intersection = mask[None, :] & masks[candidate_mask]
        joint = mask[None, :] | masks[candidate_mask]

        intersection_areas = np.sum(intersection, axis=(1, 2))
        joint_areas = np.sum(joint, axis=(1, 2))

        relative_intersections = intersection_areas / joint_areas

        over_threshold = np.where(relative_intersections > mask_iou_threshold)

        higher_scored = scores[candidate_mask][over_threshold] > score

        drop = np.sum(higher_scored) >= 1

        if drop:
            filter_mask[i] = False

        # print(relative_intersections)

    return np.arange(len(masks))[filter_mask]


def prediction(image, model, min_score=0.0, tiling=None):
    # apply tiled inference
    if tiling:
        all_detections = tiled_inference(image, model, **tiling)
        # filter by score
        all_detections = list(
            filter(lambda det: det["score"] > min_score, all_detections)
        )
        # perform non-max supressions (due to tiling this is needed)
        if len(all_detections) > 0:
            filter_mask = torch_mask_nms(
                np.stack([det["mask"] for det in all_detections]),
                np.stack([det["bbox"] for det in all_detections]),
                np.stack([det["score"] for det in all_detections]),
                score_threshold=min_score,
                mask_iou_threshold=0.6,
            )

            all_detections = list(np.array(all_detections)[filter_mask])
    else:
        all_detections = inference(image, model)
        all_detections = list(
            filter(lambda det: det["score"] > min_score, all_detections)
        )

    return all_detections



"""Module for general visualization functionality
"""

from __future__ import annotations

import logging
from pathlib import Path

import cv2
import moviepy.editor as mpy
import networkx as nx
import numpy as np
from matplotlib import font_manager
from PIL import Image, ImageDraw, ImageFont
from tqdm.auto import tqdm

from acia import ureg
from acia.base import BaseImage, ImageSequenceSource, Overlay
from acia.segm.local import InMemorySequenceSource, LocalImage

# loda the deja vu sans default font
default_font = font_manager.findfont("DejaVu Sans")


def draw_scale_bar(
    image_iterator,
    xy_position: tuple[int, int],
    size_of_pixel,
    bar_width,
    bar_height,
    color=(255, 255, 255),
    font_size=25,
    font_path=default_font,
    background_color=None,
    background_margin_pixel=3,
):
    """Draws a scale bar on all images of an image sequence or iterable image array

    Args:
        image_iterator: image sequence or iterator over images
        xy_position (tuple[int, int]): lower left xy position of the scale bar
        size_of_pixel (_type_): metric size of a pixel (e.g. 0.007 * ureg.micrometer)
        bar_width (_type_): width of the scale bar (e.g. 5 * ureg.micrometer)
        short_title (str, optional): Short title of the unit to be displayed. Defaults to "μm".
        color (tuple, optional): Color of scale bar and text. Defaults to (255, 255, 255).
        font_size (int, optional): text font size. Defaults to 25.
        font_path (str, optional): text font. Defaults to "/usr/share/fonts/truetype/dejavu/DejaVuSerif.ttf".
        background_color: color for a potential background rectangle (e.g. (0, 0, 0)). Defaults to None (no background drawn).
        background_margin_pixel: pixels of margin for the background rectangle

    Yields:
        np.ndarray | LocalImage: Image in numpy format or LocalImage (depending on the input format)
    """

    # create pint quantities (values and units)
    bar_width = ureg.Quantity(bar_width)
    bar_height = ureg.Quantity(bar_height)
    size_of_pixel = ureg.Quantity(size_of_pixel)

    # load font
    font = ImageFont.truetype(font_path, font_size)

    # compute width and height of the scale bar in pixels (we need to round here)
    bar_pixel_width = int(
        np.round((bar_width / size_of_pixel).to_base_units().magnitude)
    )
    bar_pixel_height = int(
        np.round((bar_height / size_of_pixel).to_base_units().magnitude)
    )

    # extract position
    xstart, ystart = xy_position

    for image in image_iterator:

        # do we have a wrapped image?
        is_wrapped = isinstance(image, BaseImage)

        # unwrap if necessary
        if is_wrapped:
            image = image.raw

        # compute text size
        text = f"{bar_width:~P}"
        img_pil = Image.fromarray(image)
        draw = ImageDraw.Draw(img_pil)

        # get size of text
        left, top, right, bottom = draw.textbbox((xstart, ystart), text, font=font)

        text_width = right - left
        text_height = bottom - top

        if background_color:
            cv2.rectangle(
                image,
                (xstart - background_margin_pixel, ystart + background_margin_pixel),
                (
                    xstart + bar_pixel_width + background_margin_pixel,
                    ystart
                    - text_height
                    - bar_pixel_height
                    - 5
                    - background_margin_pixel,
                ),
                background_color,
                -1,
            )

        # draw scale bar
        cv2.rectangle(
            image,
            (xstart, ystart),
            (xstart + bar_pixel_width, ystart - bar_pixel_height),
            (255, 255, 255),
            -1,
        )

        img_pil = Image.fromarray(image)
        draw = ImageDraw.Draw(img_pil)

        # draw text centered and with distance to the scale bar
        draw.text(
            (
                xstart + bar_pixel_width / 2 - text_width / 2,
                ystart - text_height - bar_pixel_height - 10,
            ),
            text,
            fill=color,
            font=font,
        )

        # convert PIL image back to numpy
        image = np.array(img_pil)

        # do the image wrapping
        if is_wrapped:
            yield LocalImage(image)
        else:
            yield image


def draw_time(
    image_iterator,
    xy_position,
    time_step,
    color=(255, 255, 255),
    font_size=25,
    font_path=default_font,
    background_color=None,
    background_margin_pixel=3,
):
    """Draw time onto images

    Args:
        image_iterator (_type_): image sequence or iterator over images
        xy_position (tuple[int, int]): lower left xy position of the time text
        time_step (_type_): time step between images (e.g. 15 * ureg.minute or "15 minute")
        color (_type): Color of the time text. Defaults to (255, 255, 255) which is white.
        font_size (int, optional): text font size. Defaults to 25.
        font_path (str, optional): text font. Defaults to "/usr/share/fonts/truetype/dejavu/DejaVuSerif.ttf".
        background_color: color for a potential background rectangle (e.g. (0, 0, 0)). Defaults to None (no background drawn).
        background_margin_pixel: pixels of margin for the background rectangle

    Yields:
        _type_: _description_
    """

    time_step = ureg.Quantity(time_step)

    # load font
    font = ImageFont.truetype(font_path, font_size)

    for frame, image in enumerate(image_iterator):

        # do we have a wrapped image?
        is_wrapped = isinstance(image, BaseImage)

        # unwrap if necessary
        if is_wrapped:
            image = image.raw

        # convert to pillow image
        pil_image = Image.fromarray(image)
        draw = ImageDraw.Draw(pil_image)

        # extract time in hours and minutes
        time = (frame * time_step).to(ureg.hour)
        hours = int(np.floor(time.magnitude))
        minutes = int(np.round((time - hours * ureg.hour).to("minute").magnitude))

        time_text = f"Time: {hours:2d}:{minutes:02d} h"

        if background_color:
            # get size of text
            left, top, right, bottom = draw.textbbox(xy_position, time_text, font=font)

            text_width = right - left
            text_height = bottom - top

            x, y = xy_position

            cv2.rectangle(
                image,
                (x - background_margin_pixel, y - background_margin_pixel),
                (
                    x + text_width + background_margin_pixel,
                    y + text_height + background_margin_pixel + 5,
                ),
                background_color,
                -1,
            )

            # convert to pillow image
            pil_image = Image.fromarray(image)
            draw = ImageDraw.Draw(pil_image)

        # draw on image
        draw.text(xy_position, time_text, fill=color, font=font)

        # convert PIL image back to numpy
        image = np.array(pil_image)

        # do the image wrapping
        if is_wrapped:
            yield LocalImage(image)
        else:
            yield image


class VideoExporter:
    """
    Wrapper for opencv video writer. Simplifies usage
    """

    def __init__(self, filename, framerate, codec="MJPG"):
        self.filename = filename
        self.framerate = framerate
        self.out = None
        self.frame_height = None
        self.frame_width = None
        self.codec = codec

    def __del__(self):
        if self.out:
            self.close()

    def write(self, image):
        height, width = image.shape[:2]
        if self.out is None:
            self.frame_height, self.frame_width = image.shape[:2]
            self.out = cv2.VideoWriter(
                self.filename,
                cv2.VideoWriter_fourcc(*self.codec),
                self.framerate,
                (self.frame_width, self.frame_height),
            )
        if self.frame_height != height or self.frame_width != width:
            logging.warning(
                "You add images of different resolution to the VideoExporter. This may cause problems (e.g. black video output)!"
            )
        self.out.write(image)

    def close(self):
        if self.out:
            self.out.release()
            self.out = None

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        if self.out is None:
            logging.warning(
                "Closing video writer without any images written and no video output generated! Did you forget to write the images="
            )
        self.close()


class VideoExporter2:
    """
    Wrapper for opencv video writer. Simplifies usage
    """

    def __init__(
        self, filename: Path, framerate: int, codec="mjpeg", ffmpeg_params=None
    ):
        self.filename = Path(filename)
        self.framerate = framerate
        self.codec = codec

        if ffmpeg_params is None:
            ffmpeg_params = []

        self.ffmpeg_params = ffmpeg_params

        self.images = []

    @staticmethod
    def default_vp9(
        filename: Path,
        framerate: int,
    ):
        ffmpeg_params = ["-crf", "30", "-b:v", "0", "-speed", "1"]
        return VideoExporter2(
            filename, framerate, codec="libvpx-vp9", ffmpeg_params=ffmpeg_params
        )

    @staticmethod
    def fast_vp9(
        filename: Path,
        framerate: int,
    ):
        ffmpeg_params = ["-crf", "35", "-b:v", "0", "-speed", "3"]
        return VideoExporter2(
            filename, framerate, codec="libvpx-vp9", ffmpeg_params=ffmpeg_params
        )

    @staticmethod
    def default_h264(
        filename: Path,
        framerate: int,
    ):
        ffmpeg_params = ["-crf", "30", "-preset", "fast"]
        return VideoExporter2(
            filename, framerate, codec="libx264", ffmpeg_params=ffmpeg_params
        )

    @staticmethod
    def default_h265(filename: Path, framerate: int):
        ffmpeg_params = ["-crf", "26", "-preset", "fast"]
        return VideoExporter2(
            filename, framerate, codec="libx265", ffmpeg_params=ffmpeg_params
        )

    @staticmethod
    def default_mjpg(filename: Path, framerate: int):
        ffmpeg_params = []
        return VideoExporter2(
            filename, framerate, codec="mjpeg", ffmpeg_params=ffmpeg_params
        )

    # av1 not yet supported
    #    @staticmethod
    #    def default_av1(filename: Path, framerate: int, ffmpeg_params=["-crf", "26", "-preset", "2", "-strict", "2"]):
    #        return VideoExporter2(filename, framerate, codec="libaom-av1", ffmpeg_params=ffmpeg_params)

    def write(self, image):
        self.images.append(image)

    def close(self):
        if len(self.images) == 0:
            logging.warning(
                "Closing video writer without any images written and no video output generated! Did you forget to write the images?"
            )
        else:
            # do the video rendering
            clip = mpy.ImageSequenceClip(
                list(self.images),
                fps=self.framerate,
            )
            clip.write_videofile(
                str(self.filename.absolute()),
                codec=self.codec,
                ffmpeg_params=self.ffmpeg_params,
                # verbose=False,
                # logger=None,
            )
            self.images = []

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()


def render_segmentation(
    imageSource: ImageSequenceSource,
    overlay: Overlay = None,
    cell_color=(255, 255, 0),
) -> ImageSequenceSource:
    """Render a video of the time-lapse including the segmentaiton information.

    Args:
        imageSource (ImageSequenceSource): Your time-lapse source object.
        Overlay ([type]): Your source of RoIs for the image (e.g. cells). If None, no RoIs are visualized. Defaults to None.
        cell_color: rgb color of the cell outlines
    """

    if overlay is None:
        # when we have no rois -> create iterator that always returns None
        def always_none():
            while True:
                yield None

        overlay = iter(always_none())

    images = []

    for image, frame_overlay in tqdm(zip(imageSource, overlay.timeIterator())):
        # extract the numpy image
        if isinstance(image, BaseImage):
            image = image.raw
        elif isinstance(image, np.ndarray):
            pass
        else:
            raise Exception("Unsupported image type!")

        # copy image as we draw onto it
        image = np.copy(image)

        # Draw overlay
        if frame_overlay:
            image = frame_overlay.draw(image, cell_color)  # RGB format

        images.append(image)

    # return as sequence source again
    return InMemorySequenceSource(np.stack(images))


def render_tracking(
    image_source: ImageSequenceSource,
    overlay: Overlay,
    tracking_graph: nx.DiGraph,
) -> ImageSequenceSource:
    """Render the tracking to an image source

    Args:
        image_source (ImageSequenceSource): Image source
        overlay (Overlay): overla of cell detections (for center points)
        tracking_graph (nx.DiGraph): the tracking graph where every cell detection is a node in the graph.

    Returns:
        ImageSequenceSource: Rendered image source
    """

    images = []

    contour_lookup = {cont.id: cont for cont in overlay}

    for image, frame_overlay in tqdm(zip(image_source, overlay.timeIterator())):

        np_image = np.copy(image.raw)

        for cont in frame_overlay:
            if cont.id in tracking_graph.nodes:
                edges = tracking_graph.out_edges(cont.id)

                born = tracking_graph.in_degree(cont.id) == 0

                for edge in edges:
                    source = contour_lookup[edge[0]].center
                    target = contour_lookup[edge[1]].center

                    line_color = (255, 0, 0)  # rgb: red

                    if len(edges) > 1:
                        line_color = (0, 0, 255)  # bgr: blue

                    cv2.line(
                        np_image,
                        tuple(map(int, source)),
                        tuple(map(int, target)),
                        line_color,
                        thickness=3,
                    )

                    if born:
                        cv2.circle(
                            np_image,
                            tuple(map(int, source)),
                            3,
                            (203, 192, 255),
                            thickness=1,
                        )

                if len(edges) == 0:
                    cv2.rectangle(
                        np_image,
                        np.array(cont.center).astype(np.int32) - 2,
                        np.array(cont.center).astype(np.int32) + 2,
                        (203, 192, 255),
                    )

        images.append(np_image)

    return InMemorySequenceSource(images)


def render_video(
    image_source: ImageSequenceSource,
    filename: str,
    framerate: int,
    codec: str,
) -> None:
    """Render video

    Args:
        image_source (ImageSequenceSource): sequence of images
        filename (str): video filename
        framerate (int): framerate of the video
        codec (str): the codec for video encoding
    """

    with VideoExporter2(filename, framerate=framerate, codec=codec) as ve:
        for im in image_source:
            ve.write(im.raw)



#!/usr/bin/env python

"""The setup script."""

from setuptools import find_packages, setup

with open("README.md", encoding="utf-8") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst", encoding="utf-8") as history_file:
    history = history_file.read()


with open("requirements.txt", encoding="utf-8") as req_file:
    requirements = req_file.read().splitlines()

test_requirements = [
    "pytest>=3",
]

setup(
    author="Johannes Seiffarth",
    author_email="j.seiffarth@fz-juelich.de",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="The acia library provides utility functionality for analyzing 2D+t time-lapse image sequences in microfluidic live-cell imaging experiments.",
    entry_points={
        "console_scripts": [
            "acia=acia.cli:main",
        ],
    },
    install_requires=requirements,
    extras_require={
        "omero": ["omero-py>=5.9.3"],
    },
    license="MIT license",
    long_description=readme,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords="acia",
    name="acia",
    packages=find_packages(include=["acia", "acia.*"]),
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/JojoDevel/acia",
    version="0.2.37",
    zip_safe=False,
)



"""Global utilities
"""

from __future__ import annotations

import numpy as np
import rasterio
import shapely
from rasterio import features
from shapely.geometry import MultiPolygon, Polygon


def lut_mapping(image, in_min, in_max, out_min, out_max, dtype=None):
    mapped_data = (np.clip(image, in_min, in_max) - in_min) / (in_max - in_min) * (
        out_max - out_min
    ) + out_min

    if dtype:
        mapped_data = mapped_data.astype(dtype)

    return mapped_data


def pairwise_distances(points: np.ndarray):
    distances = []

    if len(points) == 0:
        return distances

    for a, b in zip(points, points[1:]):
        distances.append(np.linalg.norm(a - b))

    return distances


def mask_to_polygons(mask: np.ndarray) -> Polygon | MultiPolygon:
    """Convert a mask to a Polygon or Multipolygon

    Args:
        mask (np.ndarray): Binary mask for an object

    Returns:
        shapely.geometry.Polygon | shapely.geometry.MultiPolygon: Extracted polygon structure
    """
    all_polygons = []
    for shape, _ in features.shapes(mask.astype(np.int16), mask=(mask > 0)):
        all_polygons.append(shapely.geometry.shape(shape))

    if len(all_polygons) > 1:
        all_polygons = shapely.geometry.MultiPolygon(all_polygons)
    else:
        all_polygons = all_polygons[0]

    if not all_polygons.is_valid:
        all_polygons = all_polygons.buffer(0)

    return all_polygons


def multi_mask_to_polygons(
    mask: np.ndarray,
) -> list[tuple[int, Polygon | MultiPolygon]]:
    unique_values = np.unique(mask)
    instance_ids = unique_values[unique_values > 0]

    polygons = []

    for instance_id in instance_ids:
        polygons.append((instance_id, mask_to_polygons(mask == instance_id)))

    return polygons


def polygon_to_mask(polygon, height: int, width: int):
    """Converts a polygon to a mask

    Args:
        polygon (_type_): shapely polygon or multipolygon
        height (int): height of the mask
        width (int): width of the mask

    Returns:
        (np.ndarray): boolean mask
    """
    return rasterio.features.rasterize(
        [polygon],
        out_shape=(height, width),
    ).astype(bool)


class ScaleBar:
    """Scalebar class"""

    def draw(self, image: np.ndarray, xstart: int, ystart: int):
        raise NotImplementedError("Do not use the base class")
